@@ -1,75 +1,38 @@
 'use strict';
-var fs = require('fs-ext');
-var os = require('os');
-var sleep = require('sleep');
+const fs = require('fs-ext');
+const os = require('os');
+const sleep = require('sleep');
 
-var tty = require('tty');
+const tty = require('tty');
 
-var Hash = require('./hash.js');
+const Hash = require('./hash.js');
 
-var core = require('./core.js');
+const core = require('./core.js');
 
-var child_process = require('child_process');
+const hll = require('./hll.js');
 
-var NQPObject = require('./nqp-object.js');
+const child_process = require('child_process');
 
-var mkdirp = require('mkdirp');
+const NQPObject = require('./nqp-object.js');
+
+const mkdirp = require('mkdirp');
 
 const NQPException = require('./nqp-exception.js');
 
-var nqp = require('nqp-runtime');
+const nqp = require('./runtime.js');
 
-class SyncPipe extends NQPObject {
-  $$eoffh() {
-    if (this.$$buffer) {
-      return (this.$$buffer.length ? 0 : 1);
-    } else {
-      throw new NQPException(`Can't use eoffh, this syncpipe is not connected yet`);
-    }
-  }
+const Null = require('./null.js');
 
-  /* TODO: optimize to use the lowlevel Buffer inside the highlevel one without copying */
+const NQPInt = require('./nqp-int.js');
+const NQPStr = require('./nqp-str.js');
 
-  $$readfh(buf, size) {
-    if (!this.$$buffer) {
-      throw new NQPException(`Can't use readfh, this syncpipe is not connected yet`);
-    }
-
-    let lowlevel = this.$$buffer.slice(0, size);
-    this.$$buffer = this.$$buffer.slice(size);
-
-    let elementSize = core.byteSize(buf);
-
-    let isUnsigned = buf._STable.REPR.type._STable.REPR.isUnsigned;
-
-    if (lowlevel) {
-      let offset = 0;
-      buf.array.length = lowlevel.length / elementSize;
-      for (var i = 0; i < lowlevel.length / elementSize; i++) {
-        buf.array[i] = isUnsigned ? lowlevel.readUIntLE(offset, elementSize) : lowlevel.readIntLE(offset, elementSize);
-        offset += elementSize;
-      }
-    } else {
-      buf.array.length = 0;
-    }
-
-    return buf;
-  }
-
-  $$closefh_i() {
-    return this.$$status;
-  }
-
-  $$can() {
-    return 0;
-  }
-};
+const BOOT = require('./BOOT.js');
 
 function boolish(bool) {
   return bool ? 1 : 0;
 }
 
-var op = {};
+const op = {};
 exports.op = op;
 
 op.print = function(arg) {
@@ -78,12 +41,7 @@ op.print = function(arg) {
 };
 
 op.say = function(arg) {
-  if (process.stdout) {
-    process.stdout.write(arg);
-    process.stdout.write('\n');
-  } else {
-    console.log(arg);
-  }
+  process.stdout.write(arg + '\n');
   return arg;
 };
 
@@ -131,11 +89,12 @@ function stat(file, code, lstat) {
   const PLATFORM_BLOCKS = -7;
 
   // we can't use fs.existsSync(file) as it follows symlinks
+  let stats;
   try {
     if (lstat || code == ISLNK) {
-      var stats = fs.lstatSync(file);
+      stats = fs.lstatSync(file);
     } else {
-      var stats = fs.statSync(file);
+      stats = fs.statSync(file);
     }
   } catch (err) {
     if (code == EXISTS && err.code === 'ENOENT') {
@@ -191,6 +150,10 @@ op.lstat_time = function(file, code) {
 };
 
 class IOHandle extends NQPObject {
+  $$can(ctx, name) {
+    return 0;
+  }
+
   $$toBool(ctx) {
     return 1;
   }
@@ -207,8 +170,8 @@ class IOHandle extends NQPObject {
     this.seps = seps.array;
   }
 
-  $$setencoding(encoding) {
-    this.encoding = core.renameEncoding(encoding);
+  $$setbuffersizefh(size) {
+    return this;
   }
 };
 
@@ -216,11 +179,12 @@ class FileHandle extends IOHandle {
   constructor(fd) {
     super();
     this.fd = fd;
+    this.eof = false;
   }
 
-  $$closefh_i() {
+  $$closefh() {
     fs.closeSync(this.fd);
-    return 0;
+    return this;
   }
 
   $$isttyfh() {
@@ -232,20 +196,12 @@ class FileHandle extends IOHandle {
     return 1;
   }
 
-  $$can(ctx, name) {
-    return 0;
-  }
-
   $$filenofh() {
     return this.fd;
   }
 
   $$eoffh() {
-    // I haven't found a way to implement this directly in node.js
-    var current = fs.seekSync(this.fd, 0, 1);
-    var end = fs.seekSync(this.fd, 0, 2);
-    fs.seekSync(this.fd, current, 0);
-    return current == end ? 1 : 0;
+    return this.eof ? 1 : 0;
   }
 
   $$tellfh() {
@@ -268,18 +224,17 @@ class FileHandle extends IOHandle {
   }
 
   $$writefh(buf) {
-    var buffer = core.toRawBuffer(buf);
+    const buffer = core.toRawBuffer(buf);
     return fs.writeSync(this.fd, buffer, 0, buffer.length);
   }
 
-  $$setencoding(encoding) {
-    this.encoding = core.renameEncoding(encoding);
-  }
-
   $$readfh(buf, bytes) {
-    let isUnsigned = buf._STable.REPR.type._STable.REPR.isUnsigned;
-    let buffer = Buffer.allocUnsafe(bytes);
-    let read = fs.readSync(this.fd, buffer, 0, bytes, null);
+    const isUnsigned = buf.$$STable.REPR.type.$$STable.REPR.isUnsigned;
+    const buffer = Buffer.allocUnsafe(bytes);
+    const read = fs.readSync(this.fd, buffer, 0, bytes, null);
+    if (read === 0) {
+      this.eof = true;
+    }
     buf.array.length = read;
     for (let i = 0; i < read; i++) {
       if (isUnsigned) {
@@ -307,7 +262,7 @@ function modeToFlags(mode) {
     else throw 'unknown mode to open: ' + mode;
   }
 
-  for (let c of mode.substr(1)) {
+  for (const c of mode.substr(1)) {
     if (c === 'a') flags |= fs.constants.O_APPEND;
     else if (c === 'c') flags |= fs.constants.O_CREAT;
     else if (c === 't') flags |= fs.constants.O_TRUNC;
@@ -319,30 +274,28 @@ function modeToFlags(mode) {
 }
 
 op.open = function(name, mode) {
-  let fh = new FileHandle(fs.openSync(name, modeToFlags(mode)));
-  fh.encoding = 'utf8';
-  return fh;
+  try {
+    const fh = new FileHandle(fs.openSync(name, modeToFlags(mode)));
+    return fh;
+  } catch (e) {
+    if (e.code === 'ENOENT') {
+      throw new NQPException(`Failed to open file ${name}: No such file or director`);
+    } else {
+      throw e;
+    }
+  }
 };
 
 op.seekfh = function(ctx, fh, offset, whence) {
   if (whence == 0 && offset < 0) {
-    ctx.die(`Can't seek to position: ${offset}`);
+    throw new NQPException(`Can't seek to position: ${offset}`);
   }
   if (!(whence == 0 || whence == 1 || whence == 2)) {
-    ctx.die('Invalid whence passed to seekfh: ' + whence);
+    throw new NQPException('Invalid whence passed to seekfh: ' + whence);
   }
   return fs.seekSync(fh.fd, offset, whence);
 };
 
-op.closefh = function(fh) {
-  fh.$$closefh_i();
-};
-
-op.closefh_i = function(fh) {
-  return fh.$$closefh_i();
-};
-
-
 op.unlink = function(filename) {
   try {
     fs.unlinkSync(filename);
@@ -366,7 +319,16 @@ op.chdir = function(dir) {
 };
 
 op.rmdir = function(dir) {
-  fs.rmdirSync(dir);
+  try {
+    fs.rmdirSync(dir);
+  } catch (err) {
+    if (err.code === 'ENOENT') {
+    } else if (err.code === 'ENOTEMPTY') {
+      throw new NQPException('Failed to rmdir: directory not empty');
+    } else {
+      throw err;
+    }
+  }
 };
 
 op.mkdir = function(dir, mode) {
@@ -390,96 +352,28 @@ class DirHandle {
 
   $$closedir() {
   }
+
+  $$istype(type) {
+    return 0;
+  }
+
+  $$toBool(ctx) {
+    return 1;
+  }
 };
 
 op.opendir = function(path) {
   return new DirHandle(fs.readdirSync(path));
 };
 
-const PIPE_INHERIT = 1;
-const PIPE_IGNORE = 2;
-const PIPE_CAPTURE = 4;
-const PIPE_INHERIT_IN = 1;
-const PIPE_IGNORE_IN = 2;
-const PIPE_CAPTURE_IN = 4;
-const PIPE_INHERIT_OUT = 8;
-const PIPE_IGNORE_OUT = 16;
-const PIPE_CAPTURE_OUT = 32;
-const PIPE_INHERIT_ERR = 64;
-const PIPE_IGNORE_ERR = 128;
-const PIPE_CAPTURE_ERR = 256;
-
-function stringifyEnv(ctx, hash) {
-  let stringifed = {};
-
-  hash.content.forEach(function(value, key, map) {
-    stringifed[key] = nqp.toStr(value, ctx);
-  });
-
-  return stringifed;
-}
-
-function stringifyArray(ctx, array) {
-  let stringified = [];
-  for (let element of array.array) {
-    stringified.push(nqp.toStr(element, ctx));
-  }
-  return stringified;
-}
-
-op.syncpipe = function() {
-  return new SyncPipe();
-};
-
-function run(isShell, ctx, command, dir, env, input, output, error, flags) {
-  const options = {
-    shell: isShell,
-    cwd: dir,
-    env: stringifyEnv(ctx, env),
-    stdio: [process.stdin, 'pipe', 'pipe'],
-  };
-
-  let result;
-  if (isShell) {
-    result = child_process.spawnSync(command, options);
-  } else {
-    let stringified = stringifyArray(ctx, command);
-    result = child_process.spawnSync(stringified.shift(), stringified, options);
-  }
-
-  if (flags & PIPE_CAPTURE_IN) {
-    throw new NQPException('nqp::shell with PIPE_CAPTURE_IN NYI');
-  }
-
-  if (flags & PIPE_CAPTURE_OUT) {
-    output.$$buffer = result.output[1];
-    output.$$status = result.status << 8;
-  }
-
-  if (flags & PIPE_CAPTURE_ERR) {
-    error.$$buffer = result.output[2];
-    error.$$status = result.status << 8;
-  }
-
-  return result.status << 8;
-};
-
-op.shell = function(ctx, command, dir, env, input, output, error, flags) {
-  return run(true, ctx, command, dir, env, input, output, error, flags);
-};
-
-op.spawn = function(ctx, command, dir, env, input, output, error, flags) {
-  return run(false, ctx, command, dir, env, input, output, error, flags);
-};
-
 op.cwd = function() {
   return process.cwd();
 };
 
 op.getenvhash = function() {
-  var hash = new Hash();
-  for (var key in process.env) {
-    hash.content.set(key, process.env[key]);
+  const hash = new Hash();
+  for (const key in process.env) {
+    hash.content.set(key, new NQPStr(process.env[key]));
   }
   return hash;
 };
@@ -487,7 +381,6 @@ op.getenvhash = function() {
 class StdHandle extends IOHandle {
   constructor() {
     super();
-    this.encoding = 'utf8';
   }
 };
 
@@ -505,17 +398,23 @@ class Stderr extends StdHandle {
   }
 
   $$writefh(buf) {
-    var buffer = core.toRawBuffer(buf);
+    const buffer = core.toRawBuffer(buf);
     process.stderr.write(buffer);
   }
 };
 
+const STDERR = new Stderr();
 op.getstderr = function() {
-  return new Stderr();
+  return STDERR;
 };
 
 class Stdout extends StdHandle {
-  isttyfh() {
+  constructor() {
+    super();
+    this.tellCount = 0;
+  }
+
+  $$isttyfh() {
     return (process.stdout.isTTY ? 1 : 0);
   }
 
@@ -527,28 +426,37 @@ class Stdout extends StdHandle {
   }
 
   $$writefh(buf) {
-    var buffer = core.toRawBuffer(buf);
+    const buffer = core.toRawBuffer(buf);
+    this.tellCount += buffer.length;
     process.stdout.write(buffer);
   }
+
+  /* HACK - needed for the REPL to run */
+  $$tellfh() {
+    return this.tellCount;
+  }
 };
 
+const STDOUT = new Stdout();
 op.getstdout = function() {
-  return new Stdout();
+  return STDOUT;
 };
 
 
-class Stdin extends StdHandle {
-  $$filenofh() {
-    return process.stdin.fd;
-  }
-
-  $$isttyfh() {
-    return (process.stdin.isTTY ? 1 : 0);
-  }
-};
-
+let stdin;
 op.getstdin = function() {
-  return new Stdin();
+  if (!stdin) {
+    let fd;
+    try {
+      fd = fs.openSync('/dev/stdin', 'rs');
+    } catch (e) {
+      /* this should work on Windows, we need to test it tho */
+      fd = 0;
+    };
+    stdin = new FileHandle(fd);
+  }
+
+  return stdin;
 };
 
 op.exit = function(code) {
@@ -580,3 +488,141 @@ op.rename = function(oldPath, newPath) {
   fs.renameSync(oldPath, newPath);
 };
 
+function wrapBuffer(buffer, type) {
+  const elementSize = core.byteSize(type);
+
+  const isUnsigned = type.$$STable.REPR.type.$$STable.REPR.isUnsigned;
+
+  const wrapped = type.$$STable.REPR.allocate(type.$$STable);
+
+  let offset = 0;
+  for (let i = 0; i < buffer.length / elementSize; i++) {
+    wrapped.array[i] = isUnsigned ? buffer.readUIntLE(offset, elementSize) : buffer.readIntLE(offset, elementSize);
+    offset += elementSize;
+  }
+
+  return wrapped;
+}
+
+/*async*/ function stringifyEnv(ctx, hash) {
+  const stringifed = {};
+
+  for (const key of hash.content.keys()) {
+    stringifed[key] = /*await*/ nqp.toStr(hash.content.get(key), ctx);
+  }
+
+  return stringifed;
+}
+
+/*async*/ function stringifyArray(ctx, array) {
+  const stringified = [];
+  for (const element of array.array) {
+    stringified.push(/*await*/ nqp.toStr(element, ctx));
+  }
+  return stringified;
+}
+
+op.spawnprocasync = /*async*/ function(ctx, queue, args, cwd, env, config) {
+  const options = {
+    shell: false,
+    cwd: cwd,
+    env: /*await*/ stringifyEnv(ctx, env),
+    stdio: [
+      process.stdin,
+      config.content.get('stdout_bytes') ? 'pipe' : process.stdout,
+      config.content.get('stderr_bytes') ? 'pipe' : process.stderr,
+    ],
+  };
+
+  const stringified = /*await*/ stringifyArray(ctx, args);
+
+  const result = child_process.spawnSync(stringified.shift(), stringified, options);
+
+  if (config.content.get('ready')) {
+    /*await*/ config.content.get('ready').$$call(ctx, null);
+  }
+
+
+  let str_box = ctx.$$getHLL().get('str_box');
+
+  if (str_box === undefined) str_box = Null;
+
+  if (config.content.get('stdout_bytes')) {
+    /*await*/ config.content.get('stdout_bytes').$$call(ctx, null, new NQPInt(0), wrapBuffer(result.output[1], config.content.get('buf_type')), str_box);
+    /*await*/ config.content.get('stdout_bytes').$$call(ctx, null, new NQPInt(1), str_box, str_box);
+  }
+
+  if (config.content.get('stderr_bytes')) {
+    /*await*/ config.content.get('stderr_bytes').$$call(ctx, null, new NQPInt(0), wrapBuffer(result.output[2], config.content.get('buf_type')), str_box);
+    /*await*/ config.content.get('stderr_bytes').$$call(ctx, null, new NQPInt(1), str_box, str_box);
+  }
+
+  if (config.content.get('done')) {
+    /*await*/ config.content.get('done').$$call(ctx, null, new NQPInt(result.status << 8));
+  }
+};
+
+op.permit = function(handle, channel, permits) {
+  // TODO Implement permit handling properly
+  return handle;
+};
+
+let sigCache = null;
+op.getsignals = function(currentHLL) {
+  if (sigCache) {
+    return sigCache;
+  }
+
+  const osSigs = os.platform() === 'win32'
+    ? {
+        SIGHUP: 1,
+        SIGKILL: 9,
+        SIGWINCH: 28,
+      }
+    : os.constants.signals;
+
+  const sigWanted = [
+    'SIGHUP', 'SIGINT', 'SIGQUIT', 'SIGILL', 'SIGTRAP', 'SIGABRT',
+    'SIGEMT', 'SIGFPE', 'SIGKILL', 'SIGBUS', 'SIGSEGV', 'SIGSYS',
+    'SIGPIPE', 'SIGALRM', 'SIGTERM', 'SIGURG', 'SIGSTOP', 'SIGTSTP',
+    'SIGCONT', 'SIGCHLD', 'SIGTTIN', 'SIGTTOU', 'SIGIO', 'SIGXCPU',
+    'SIGXFSZ', 'SIGVTALRM', 'SIGPROF', 'SIGWINCH', 'SIGINFO', 'SIGUSR1',
+    'SIGUSR2', 'SIGTHR', 'SIGSTKFLT', 'SIGPWR', 'SIGBREAK',
+  ];
+
+  const arr = [];
+  for (const k of sigWanted) {
+    arr.push( core.strToObj(currentHLL, k), core.intToObj(currentHLL, k in osSigs ? osSigs[k] : 0) );
+  }
+
+  const res = hll.slurpyArray(currentHLL, arr);
+  sigCache = res;
+  return res;
+};
+
+op.getpid = function() {
+  return process.pid;
+};
+
+op.getppid = function() {
+  return process.ppid;
+};
+
+function uname(flag) {
+  const result = child_process.spawnSync('uname', [flag]);
+  if (result.error) {
+    return "";
+  } else {
+    return result.stdout.toString().replace(/\n$/, '');
+  }
+}
+
+op.uname = function() {
+  let sysname = os.type();
+  let release = os.release();
+  let version = uname('-v');
+  let machine = uname('-m');
+
+  const stable = BOOT.StrArray.$$STable;
+  return stable.REPR.allocateFromArray(stable, [sysname, release, version, machine]);
+};
